use super::websocket;
use actix::*;
use actix_web::{ws, Error, HttpRequest, HttpResponse};
use std::time::{Duration, Instant};

const HEARTBEAT_INTERVAL: Duration = Duration::from_secs(5);
const CLIENT_TIMEOUT: Duration = Duration::from_secs(10);

pub struct WSSessionState {
    pub addr: Addr<websocket::WSServer>,
}

pub fn ws_route(req: &HttpRequest<WSSessionState>) -> Result<HttpResponse, Error> {
    ws::start(req, WSSession { hb: Instant::now() })
}

struct WSSession {
    hb: Instant,
}

impl Actor for WSSession {
    type Context = ws::WebsocketContext<Self, WSSessionState>;

    fn started(&mut self, ctx: &mut Self::Context) {
        self.hb(ctx);
        ctx.state()
            .addr
            .send(websocket::Connect {
                addr: ctx.address().recipient(),
            })
            .into_actor(self)
            .then(|res, _, ctx| {
                match res {
                    Ok(_) => (),
                    _ => ctx.stop(),
                }
                fut::ok(())
            })
            .wait(ctx);
    }

    fn stopping(&mut self, ctx: &mut Self::Context) -> Running {
        ctx.state().addr.do_send(websocket::Disconnect {
            addr: ctx.address().recipient(),
        });
        Running::Stop
    }
}

impl Handler<websocket::Message> for WSSession {
    type Result = ();

    fn handle(&mut self, msg: websocket::Message, ctx: &mut Self::Context) {
        ctx.text(msg.0);
    }
}

impl StreamHandler<ws::Message, ws::ProtocolError> for WSSession {
    fn handle(&mut self, msg: ws::Message, ctx: &mut Self::Context) {
        match msg {
            ws::Message::Ping(msg) => {
                self.hb = Instant::now();
                ctx.pong(&msg);
            }
            ws::Message::Pong(_) => {
                self.hb = Instant::now();
            }
            ws::Message::Text(_) => log::info!("Unexpected txt essage from client"),
            ws::Message::Binary(_) => log::info!("Unexpected binary message from client"),
            ws::Message::Close(_) => ctx.stop(),
        }
    }
}

impl WSSession {
    fn hb(&self, ctx: &mut ws::WebsocketContext<Self, WSSessionState>) {
        ctx.run_interval(HEARTBEAT_INTERVAL, |act, ctx| {
            if Instant::now().duration_since(act.hb) > CLIENT_TIMEOUT {
                log::info!("Websocket client timed out, disconnecting!");

                ctx.state().addr.do_send(websocket::Disconnect {
                    addr: ctx.address().recipient(),
                });
                ctx.stop();
                return;
            }
            ctx.ping("");
        });
    }
}
